//
// APPLICAZIONE BASE: SDL + OPENGL + GLSL
//////////////////////////////////////////////

// SDL headers
#include "SDL.h"

// local headers
#include "shutils.h"

// standard headers
#include <math.h>
#include <iostream>
#include <vector>

// GLEW
#include <GL/glew.h>
#include <GL/wglew.h>

#include <GL/glu.h>

static const float MOVE_X = 0.01f;
static const float MOVE_Z = 0.01f;
static const float DELTA_DEGREE = 0.03f;

// camera 
static double camx = 0.0, camy = 0.0, camz = -3.0;
static double az = 0.0, elev = 0.0;
static double tilt = 0.0;

// model rotation
float Rx = 0.0f;
float Ry = 0.0f;
float Rz = 0.0f;

// current width and height of the rendering context
static int ww, hh;

// mouse
static int start_posx;
static int start_posy;

// display adapter info
static bool flagDisplayInfo = false;
static DISPLAY_DEVICE displayDevice; 

// fps
static double fps = 0.0;
static int frame_counter = 0;
static int time_ref = 0;

/* SHADERS
/**********************************************************/

GLuint flattening_prog;
GLuint quantizecolor_prog;
GLuint current_prog;

// Compile shader and provide verbose output
void compileShader(GLuint shader)
{
	static char shlog[2048];
	GLint status;
	int length;

	glCompileShader(shader);
	glGetShaderiv(shader, GL_COMPILE_STATUS, &status);
	if (status != GL_TRUE)
	{
		glGetShaderInfoLog(shader, 2048, &length, shlog);
		std::cout << std::endl;
		std::cout << shlog << std::endl;
	}
	else
		std::cout << " OK!" << std::endl;
}

// Link shader program and provide verbose output
void linkShaderProgram(GLuint program)
{
	static char proglog[2048];
	GLint status;
	int length;

	glLinkProgram(program);
	glGetProgramiv(program, GL_LINK_STATUS, &status);
	if (status != GL_TRUE)
	{
		glGetProgramInfoLog(program, 2048, &length, proglog);
		std::cout << std::endl;
		std::cout << proglog << std::endl;
	}
	else
		std::cout << " OK!" << std::endl;
}

void setupShaders()
{
	// Color Quantization Shader
	GLuint vs1 = glCreateShader(GL_VERTEX_SHADER);
	const char *vs1_src = importShaders("shaders/quantizecolor.vert");
	glShaderSource(vs1, 1, &vs1_src, NULL);
	std::cout << "  Compile Color Quantization vertex shader...";
	compileShader(vs1);

	GLuint fs1 = glCreateShader(GL_FRAGMENT_SHADER);
	const char *fs1_src = importShaders("shaders/quantizecolor.frag");
	glShaderSource(fs1, 1, &fs1_src, NULL);
	std::cout << "  Compile Color Quantization fragment shader...";
	compileShader(fs1);

	quantizecolor_prog = glCreateProgram();
	glAttachShader(quantizecolor_prog, vs1);
	glAttachShader(quantizecolor_prog, fs1);

	std::cout << "  Link Color Quantization program...";
	linkShaderProgram(quantizecolor_prog);

	// Flatten Shader
	GLuint vs2 = glCreateShader(GL_VERTEX_SHADER);
	const char *vs2_src = importShaders("shaders/flattening.vert");
	glShaderSource(vs2, 1, &vs2_src, NULL);
	std::cout << "  Compile Flattening vertex shader...";
	compileShader(vs2);

	GLuint fs2 = glCreateShader(GL_FRAGMENT_SHADER);
	const char *fs2_src = importShaders("shaders/flattening.frag");
	glShaderSource(fs2, 1, &fs2_src, NULL);
	std::cout << "  Compile Flattening fragment shader...";
	compileShader(fs2);

	flattening_prog = glCreateProgram();
	glAttachShader(flattening_prog, vs2);
	glAttachShader(flattening_prog, fs2);

	std::cout << "  Link Flattening program...";
	linkShaderProgram(flattening_prog);

	// set current program
	current_prog = flattening_prog;
}


/* KEYBOARD HANDLING
/**********************************************************/

void processKeydownEvent(SDL_KeyboardEvent keydown)
{
	SDLKey key = keydown.keysym.sym;

	if (key == SDLK_i)
	{
		if (flagDisplayInfo) 
			flagDisplayInfo = false;
		else
			flagDisplayInfo = true;
	}
	else if (key == SDLK_1)
	{
		// standard OpenGL rendering
		current_prog = 0;
	}
	else if (key == SDLK_2)
	{
		// Flattening
		current_prog = flattening_prog;
	}
	else if (key == SDLK_3)
	{
		// Color Quantization
		current_prog = quantizecolor_prog;
	}
	else if (key == SDLK_q) 
	{
		Rx += DELTA_DEGREE;
	}
	else if (key == SDLK_w)
	{
		Rx -= DELTA_DEGREE;
	}
	else if (key == SDLK_a)
	{
		Ry += DELTA_DEGREE;
	}
	else if (key == SDLK_s)
	{
		Ry -= DELTA_DEGREE;
	}
	else if (key == SDLK_z)
	{
		Rz += DELTA_DEGREE;
	}
	else if (key == SDLK_x)
	{
		Rz -= DELTA_DEGREE;
	}
	else if (key == SDLK_UP)
	{
		camz += MOVE_Z;
	}
	else if (key == SDLK_DOWN)
	{
		camz -= MOVE_Z;
	}
	else if (key == SDLK_RIGHT)
	{
		camx -= MOVE_X;
	}
	else if (key == SDLK_LEFT)
	{
		camx += MOVE_X;
	}
}

void processKeyupEvent(SDL_KeyboardEvent keyup)
{
	//...TODO...
}

/* MOUSE HANDLING
/***************************************************************/

void processMouseButtons(SDL_MouseButtonEvent mousebutton)
{
	//mousebutton.state;
}

void processMouseMotion(SDL_MouseMotionEvent mousemotion)
{
	if (mousemotion.state == SDL_PRESSED)
	{
		double deltaAz = static_cast<double>(mousemotion.xrel) * 0.000288;
		double deltaEl = static_cast<double>(mousemotion.yrel) * 0.000288;

		az += deltaAz;
		elev += deltaEl;

		// clamping
		if (elev > 1.4)
			elev = 1.4;

		if (elev < -1.4)
			elev = -1.4;
	}
}

/* SCENE
/*********************************************************/

void getDisplayAdapterInfo()
{
	displayDevice.cb = sizeof(displayDevice);

	// Query the first display cards (id = 0)
	// (it does not work for system with multiple video cards)
	int id = 0;
	EnumDisplayDevices(NULL, id, &displayDevice, 0);

	std::cout << std::endl;
	std::cout << "  Graphics Board: " << displayDevice.DeviceString
		<< std::endl;
}

void setPerspectiveProjection()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0, 0, ww, hh);
	double aspect = static_cast<double>(ww) / static_cast<double>(hh);
	gluPerspective(45.0, aspect, 1.0, 10.0);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE); // just in case
	glDepthFunc(GL_LEQUAL);
	glShadeModel(GL_SMOOTH);
}

void resizeEvent(int w, int h)
{
	// size update
	ww = w;
	hh = h;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	setPerspectiveProjection();
}

void setOrthographicProjection() 
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();

	// set orthographic projection
	glLoadIdentity();
	gluOrtho2D(0, ww, 0, hh);

	// invert the y axis, down is positive
	glScalef(1, -1, 1);

	// mover the origin from the bottom left corner
	// to the upper left corner
	glTranslatef(0, -hh, 0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

void resetPerspectiveProjection()
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
}

void renderScene()
{
	glUseProgram(current_prog);

	glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	double lookz = cos(az) * sin(1.570 - elev);
	double lookx = sin(az) * sin(1.570 - elev);
	double looky = cos(1.570 - elev);

	gluLookAt(camx, camy, camz, 
		camx + lookx, camy + looky, camz + lookz, 
		0.0, 1.0, 0.0);

	// rotate model
	glRotatef(Rx, 1.0f, 0.0f, 0.0f);
	glRotatef(Ry, 0.0f, 1.0f, 0.0f);
	glRotatef(Rz, 0.0f, 0.0f, 1.0f);

	static GLfloat v0[] = { -1.0f, -1.0f, 1.0f };
	static GLfloat v1[] = { 1.0f, -1.0f, 1.0f };
	static GLfloat v2[] = { 1.0f, 1.0f, 1.0f };
	static GLfloat v3[] = { -1.0f, 1.0f, 1.0f };
	static GLfloat v4[] = { -1.0f, -1.0f, -1.0f };
	static GLfloat v5[] = { 1.0f, -1.0f, -1.0f };
	static GLfloat v6[] = { 1.0f, 1.0f, -1.0f };
	static GLfloat v7[] = { -1.0f, 1.0f, -1.0f };
	static GLfloat red[] = { 1.0f, 0.0f, 0.0f, 1.0f };
	static GLfloat green[] = { 0.0f, 1.0f, 0.0f, 1.0f };
	static GLfloat blue[] = { 0.0f, 0.0f, 1.0f, 1.0f };
	static GLfloat white[] = { 1.0f, 1.0f, 1.0f, 1.0f };
	static GLfloat yellow[] = { 0.0f, 1.0f, 1.0f, 1.0f };
	static GLfloat black[] = { 0.0f, 0.0f, 0.0f, 1.0f };
	static GLfloat orange[] = { 1.0f, 1.0f, 0.0f, 1.0f };
	static GLfloat purple[] = { 1.0f, 0.0f, 1.0f, 0.0f };

	glBegin(GL_TRIANGLES);
	glColor4fv( red );
	glVertex3fv( v0 );
	glColor4fv( green );
	glVertex3fv( v1 );
	glColor4fv( blue );
	glVertex3fv( v2 );
	glColor4fv( red );
	glVertex3fv( v0 );
	glColor4fv( blue );
	glVertex3fv( v2 );
	glColor4fv( white );
	glVertex3fv( v3 );
	glColor4fv( green );
	glVertex3fv( v1 );
	glColor4fv( black );
	glVertex3fv( v5 );
	glColor4fv( orange );
	glVertex3fv( v6 );
	glColor4fv( green );
	glVertex3fv( v1 );
	glColor4fv( orange );
	glVertex3fv( v6 );
	glColor4fv( blue );
	glVertex3fv( v2 );
	glColor4fv( black );
	glVertex3fv( v5 );
	glColor4fv( yellow );
	glVertex3fv( v4 );
	glColor4fv( purple );
	glVertex3fv( v7 );
	glColor4fv( black );
	glVertex3fv( v5 );
	glColor4fv( purple );
	glVertex3fv( v7 );
	glColor4fv( orange );
	glVertex3fv( v6 );
	glColor4fv( yellow );
	glVertex3fv( v4 );
	glColor4fv( red );
	glVertex3fv( v0 );
	glColor4fv( white );
	glVertex3fv( v3 );
	glColor4fv( yellow );
	glVertex3fv( v4 );
	glColor4fv( white );
	glVertex3fv( v3 );
	glColor4fv( purple );
	glVertex3fv( v7 );
	glColor4fv( white );
	glVertex3fv( v3 );
	glColor4fv( blue );
	glVertex3fv( v2 );
	glColor4fv( orange );
	glVertex3fv( v6 );
	glColor4fv( white );
	glVertex3fv( v3 );
	glColor4fv( orange );
	glVertex3fv( v6 );
	glColor4fv( purple );
	glVertex3fv( v7 );
	glColor4fv( green );
	glVertex3fv( v1 );
	glColor4fv( red );
	glVertex3fv( v0 );
	glColor4fv( yellow );
	glVertex3fv( v4 );
	glColor4fv( green );
	glVertex3fv( v1 );
	glColor4fv( yellow );
	glVertex3fv( v4 );
	glColor4fv( black );
	glVertex3fv( v5 );
	glEnd();

	SDL_GL_SwapBuffers();
	frame_counter++;
}


// check for the OpenGL extensions used in the application
void checkOpenGLExtensions()
{
	bool ok = true;

	// OpenGL driver capabilities
	std::cout << std::endl << "  Driver OpenGL: ";

	if (GLEW_VERSION_2_0)
		std::cout << "OpenGL 2.0" << std::endl;
	else if (GLEW_VERSION_1_4)
		std::cout << "OpenGL 1.4" << std::endl;
	else if (GLEW_VERSION_1_3)
		std::cout << "OpenGL 1.3" << std::endl;
	else if (GLEW_VERSION_1_2)
		std::cout << "OpenGL 1.2" << std::endl;
	else if (GLEW_VERSION_1_1)
		std::cout << "OpenGL 1.1" << std::endl;

	std::cout << std::endl;

	// Extensions

	if (GLEW_VERSION_2_0)
	{
		ok = true;
	}
	else
	{
		std::cout << std::endl;
		std::cout << "  OpenGL 2.0 is required(!!)" << std::endl << std::endl;
		exit(0);
	}

	std::cout << std::endl;

	if (!ok)
	{
		std::cout << "  One or more of the required extensions are not supported." 
			<< std::endl << std::endl;
		exit(0);
	}
	else
		std::cout << "  All the required extensions are supported (!)" << 
		std::endl << std::endl;

}

int main(int argc, char* argv[])
{
	// INIZIALIZZAZIONE SDL+OPENGL
	///////////////////////////////////////////////////////

	if( SDL_Init(SDL_INIT_VIDEO) < 0 ) 
	{
		std::cout << "Couldnt initialize SDL: " << SDL_GetError();
		exit(1);
	}

	/* Clean up on exit */
	atexit(SDL_Quit);

	ww = 500; hh = 500;

	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_Surface *screen = SDL_SetVideoMode(ww, hh, 32, SDL_OPENGL | SDL_RESIZABLE);
	if ( screen == NULL )
	{
		fprintf(stderr, "Couldnt set 640x480x8 video mode: %s\n",
			SDL_GetError());
		exit(1);
	}

	// inizializzazione GLEW
	GLenum err = glewInit();
	if (GLEW_OK != err)
	{
		std::cout << std::endl;
		std::cout << "  GLEW initialization error: " << glewGetErrorString(err) 
			<< std::endl;
		std::cout << std::endl;
		exit(-2);
	}

	std::cout << std::endl;
	std::cout << "  GLEW (version " << glewGetString(GLEW_VERSION) 
		<< ") has been initialized(!)" << std::endl; 

	// get display adapter information
	getDisplayAdapterInfo();

	// check required OpenGL extensions
	checkOpenGLExtensions();

	// setup GLSL shaders
	setupShaders();

	setPerspectiveProjection();

	// MAIN EVENT LOOP
	////////////////////////////////////////////////

	SDL_Event sdlevent;
	bool quit=false;
	while(!quit)
	{
		SDL_PollEvent(&sdlevent);

		if (sdlevent.type == SDL_VIDEORESIZE)
		{
			resizeEvent(sdlevent.resize.w, sdlevent.resize.h);
		}
		else if (sdlevent.type == SDL_KEYDOWN)
		{
			processKeydownEvent(sdlevent.key);

			if (sdlevent.key.keysym.sym == SDLK_ESCAPE)
				quit = true;
		}
		else if (sdlevent.type == SDL_KEYUP)
		{
			processKeyupEvent(sdlevent.key);
		}
		else if (sdlevent.type == SDL_MOUSEMOTION)
		{
			processMouseMotion(sdlevent.motion);
		}
		else if (sdlevent.type == SDL_MOUSEBUTTONDOWN || sdlevent.type == SDL_MOUSEBUTTONUP)
		{
			processMouseButtons(sdlevent.button);
		}

		renderScene();
	}

	return 0;
}

